#!/usr/bin/python

''' Python API for YouTube
    Copyright (C)  2013 nagev

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.  '''

__version__ = "0.3.07"
__author__ = "nagev"
__license__ = "GPLv3"

import re
import sys
import time
import json
import logging

def decode_if_py3(data):
    return data.decode("UTF-8")

if sys.version_info[:2] >= (3, 0):
    from urllib.request import build_opener
    from urllib.parse import parse_qs, unquote_plus
else:
    decode_if_py3 = lambda x: x
    from urllib2 import build_opener
    from urllib import unquote_plus
    from urlparse import parse_qs

#logging.basicConfig(level=logging.DEBUG)

def _extract_function_from_js(funcname, js):
    # Find a function definition named funcname and extract components
    match = re.search(r'function %s\(((?:\w+,?)+)\)\{([^\{]+)\}' % funcname, js)
    return {'name': funcname, 'argnames': match.group(1).split(","),
        'body': match.group(2) }

def _getval(val, argsdict): # resolves variable values. preserves int literals
    match = re.match(r'(\d+)', val)
    if match:
        return(int(match.group(1)))
    elif val in argsdict:
        return argsdict[val]
    else:
        raise RuntimeError("Error val %s from dict %s" % (val, argsdict))

def _get_func_from_call(f_old, fname, argscall, js):
    argscall = argscall.split(",")
    newfunc = _extract_function_from_js(fname, js)
    newfunc['args'] = {}
    for n, argname in enumerate(argscall):
        value = _getval(argname, f_old['args'])
        # curvar is the argument name specified in the new function definition
        curvar = newfunc['argnames'][n]
        newfunc['args'][curvar] = value
    return newfunc

def _solve(f, js):
    # solve basic javascript function
    parts = f['body'].split(";")
    for part in parts:
        logging.debug("Working on part: %s" % part)
        # split, do nothing
        m = re.match(r'(\w+)=(\w+)\.split\(""\)', part)
        if m and m.group(1) == m.group(2):
            continue
        m = re.match(r'(\w+)=(\w+)\(((?:\w+,?)+)\)', part)
        if m: # a function call
            lhs, fname, argscall = m.group(*range(1,4))
            newfunc = _get_func_from_call(f, fname, argscall, js)
            f['args'][lhs] = _solve(newfunc, js) # recursive call
            continue
        m = re.match(r'var\s(\w+)=(\w+)\[(\w+)\]', part)
        if m: # new var is an index of another var; eg: var a = b[c]
            b, c = [_getval(x, f['args']) for x in m.group(*range(2,4))]
            f['args'][m.group(1)] = b[c]
            continue
        m = re.match(r'(\w+)\[(\w+)\]=(\w+)\[(\w+)\%(\w+)\.length\]', part)
        if m: # a[b]=c[d%e.length]
            vals = m.group(*range(1,6))
            a, b, c, d, e = [_getval(x, f['args']) for x in vals]
            f['args'][m.group(1)] = a[:b] + c[d % len(e)] + a[b + 1:] 
            continue
        m = re.match(r'(\w+)\[(\w+)\]=(\w+)', part)
        if m: # a[b]=c
            a, b, c = [_getval(x, f['args']) for x in m.group(*range(1,4))]
            f['args'][m.group(1)] = a[:b] + c + a[b + 1:] # a[b] = c
            continue
        m= re.match(r'return (\w+)(\.join\(""\))?', part)
        if m: # return
            return f['args'][m.group(1)]
        m = re.match(r'(\w+)=(\w+)\.reverse\(\)', part)
        if m: # reverse        
            f['args'][m.group(1)] = _getval(m.group(2), f['args'])[::-1]
            continue
        m = re.match(r'(\w+)=(\w+)\.slice\((\w+)\)', part)
        if m:  # slice a=b.slice(c)
            a, b, c = [_getval(x, f['args']) for x in m.group(*range(1,4))]
            f['args'][m.group(1)] = b[c:]
            continue
        raise RuntimeError("no match for %s" % part)

def _decodesig(sig, js):
    # get main function name from a function call
    sigargument = "g.s"
    sigprefix = "g.sig"
    match = re.search(r'%s\|\|(\w+)\(%s\)' % (sigprefix, sigargument), js)
    funcname = match.group(1)
    function = _extract_function_from_js(funcname, js)
    if not len(function['argnames']) == 1:
        raise RuntimeError("Main sig js function has more than one arg: %s" %
            function['argnames'])
    function['args'] = {function['argnames'][0]: sig}
    return _solve(function, js)

class Stream():
    resolutions = {
        '5': ('240x400', 'flv'),
        '17': ('144x176', '3gp'),
        '18': ('360x640', 'mp4'),
        '22': ('720x1280', 'mp4'),
        '34': ('360x640', 'flv'),
        '35': ('480x854', 'flv'),
        '36': ('320x240', '3gp'),
        '37': ('1080x1920', 'mp4'),
        '38': ('3072x4096', 'superHD'),
        '43': ('360x640', 'webm'),
        '44': ('480x854', 'webm'),
        '45': ('720x1280', 'webm'),
        '46': ('1080x1920', 'webm'),
        '82': ('640x360-3D', 'mp4'),
        '84': ('1280x720-3D', 'mp4'),
        '100': ('640x360-3D', 'webm'),
        '102': ('1280x720-3D', 'webm')}
    
    def __init__(self, streammap, opener, title="ytvid", js=None):
        if not streammap.get("sig", ""):
            logging.debug("Decrypting sig: %s" % streammap['s'])
            streammap['sig'] = [_decodesig(streammap['s'][0], js)]
            logging.debug("Calculated decrypted sig: %s" % streammap['sig'][0])
        self.url = streammap['url'][0] + '&signature=' + streammap['sig'][0]
        self.vidformat = streammap['type'][0].split(';')[0]
        self.resolution = self.resolutions[streammap['itag'][0]][0]
        self.extension = self.resolutions[streammap['itag'][0]][1]
        self.itag = streammap['itag'][0]
        self.title = title
        self.filename = self.title + "." + self.extension
        self._opener = opener

    def get_filesize(self):
        opener = self._opener
        return int(opener.open(self.url).headers['content-length'])

    def download(self, progress=True, filepath=""):
        response = self._opener.open(self.url)
        total = int(response.info()['Content-Length'].strip())
        print("-Downloading '{}' [{:,} Bytes]".format(self.filename,
            total))
        status_string = ('  {:,} Bytes [{:.2%}] received. Rate: [{:4.0f} '
                         'kbps].  ETA: [{:.0f} secs]')
        chunksize, bytesdone, t0 = 1024, 0, time.time()
        outfh = open(filepath or self.filename, 'wb')
        while 1:
            chunk = response.read(chunksize)
            elapsed = time.time() - t0
            outfh.write(chunk)
            bytesdone += len(chunk)
            if not chunk:
                outfh.close()
                break
            if progress:
                rate = (bytesdone / 1024) / elapsed
                eta = (total - bytesdone) / (rate * 1024)
                display = (bytesdone, bytesdone * 1.0 / total, rate, eta)
                status = status_string.format(*display)
                sys.stdout.write("\r" + status + ' ' * 4 + "\r")
                sys.stdout.flush()
        print("\nDone")


class Pafy():

    def __len__(self):
        return self.length

    def __repr__(self):
        out = ""
        keys = "Title Author ID Duration Rating Views Thumbnail Keywords"
        keys = keys.split(" ")
        keywords = ", ".join(self.keywords)
        length = time.strftime('%H:%M:%S', time.gmtime(self.length))
        info = dict(Title=self.title,
                    Author=self.author,
                    Views=self.viewcount,
                    Rating=self.rating,
                    Duration=length,
                    ID=self.videoid,
                    Thumbnail=self.thumb,
                    Keywords=keywords)
        for k in keys: 
            try:
                out += "%s: %s\n" % (k, info[k])
            except KeyError:
                pass
        return out

    def _setmetadata(self, allinfo):
        self.title = allinfo['title'][0]
        self.author = allinfo['author'][0]
        self.videoid = allinfo['video_id'][0]
        self.rating = float(allinfo['avg_rating'][0])
        self.length = int(allinfo['length_seconds'][0])
        self.duration = time.strftime('%H:%M:%S', time.gmtime(self.length))
        self.viewcount = int(allinfo['view_count'][0])
        self.thumb = unquote_plus(allinfo['thumbnail_url'][0])
        self.formats = allinfo['fmt_list'][0].split(",")
        self.formats = [x.split("/") for x in self.formats]
        if 'keywords' in allinfo:
            self.keywords = allinfo['keywords'][0].split(',')
        if allinfo.get('iurlsd'):
            self.bigthumb = allinfo['iurlsd'][0]
        if allinfo.get('iurlmaxres'):
            self.bigthumbhd = allinfo['iurlmaxres'][0]
        return 

    def __init__(self, video_url):
        infoUrl = 'https://www.youtube.com/get_video_info?video_id='
        try:
            vidid = re.search(r'v=([a-zA-Z0-9-_]*)', video_url).group(1)
        except:
            raise RuntimeError("bad video url")
        infoUrl += vidid + "&asv=3&el=detailpage&hl=en_US"
        opener = build_opener()
        ua = ("Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64;"
              "Trident/5.0)")
        opener.addheaders = [('User-Agent', ua)]
        self.keywords = ""
        allinfo = parse_qs(decode_if_py3(opener.open(infoUrl).read()))
        self._setmetadata(allinfo)
        streamMap = allinfo['url_encoded_fmt_stream_map'][0].split(',')
        smap = [parse_qs(sm) for sm in streamMap]
        js = None
        if not smap[0].get("sig", ""):  # vevo!
            watchurl = "https://www.youtube.com/watch?v=" + vidid
            watchinfo = opener.open(watchurl).read().decode("UTF-8")
            match = re.search(r';ytplayer.config = ({.*?});', watchinfo)
            try:
                myjson = json.loads(match.group(1))
            except:
                raise RuntimeError('Problem handling this video')
            args = myjson['args']
            streamMap = args['url_encoded_fmt_stream_map'].split(",")
            html5player = myjson['assets']['js']
            js = opener.open(html5player).read().decode("UTF-8")
            smap = [parse_qs(sm) for sm in streamMap]
        self.streams = [Stream(sm, opener, self.title, js) for sm in smap]

    def getbest(self, preftype="any", ftypestrict=True):
        # set ftypestrict to False to use a non preferred format if that
        # has a higher resolution
        def _sortkey(x, key3d=0, keyres=0, keyftype=0):
            key3d = "3D" not in x.resolution
            keyres = int(x.resolution.split("x")[0])
            keyftype = preftype == x.extension
            if ftypestrict:
                return (key3d, keyftype, keyres)
            else:
                return (key3d, keyres, keyftype)
        return max(self.streams, key=_sortkey)



#!/usr/bin/python

# Pafy setup script

from distutils.core import setup

setup(
    name='Pafy',
    #packages = ['pafy'],
    py_modules=['pafy'],
    version = '0.3.07',
    description = "Python API for YouTube, query and download YouTube content",
    keywords = ["Pafy", "API", "YouTube", "youtube", "download", "video"],
    author = "nagev",
    author_email = "np1nagev@gmail.com",
    url = "http://np1.github.io/pafy/",
    download_url = "https://github.com/np1/pafy/tarball/master",
    scripts = ['ytdl'],
    package_data = {"": ["LICENSE", "README"]},
    classifiers = [
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Multimedia :: Sound/Audio :: Capture/Recording",
        "Topic :: Utilities",
        "Topic :: Multimedia :: Video",
        "Topic :: Internet :: WWW/HTTP"],
    long_description = """\

PAFY
====

Python API for YouTube

by nagev


Features:
---------

 - Download any stream for a particular video
 - Select best quality stream for download
 - Retreive metadata such as viewcount, duration, rating, author, thumbnail, keywords
 - Retrieve all availabe streams for a YouTube video (all resolutions and formats)
 - Retrieve the Download URL to download or stream the video
 - Small, standalone, single importable module file.
 - Works with age-restricted videos and non-embeddable videos
 - No dependencies
 - Command line tool (ytdl) for downloading directly from the command line

Usage Examples:
---------------

Here is how to use the module in your own python code.  For command line tool
(ytdl) instructions, see further below:

```python

>>> from pafy import Pafy
>>> url = "http://www.youtube.com/watch?v=cyMHZVT91Dw"


    # create a video instance
    
>>> video = Pafy(url)


    # get certain attributes
    
>>> video.title
u'Rick Astley Sings Live - Never Gonna Give You Up - This Morning'


>>> video.rating
4.93608852755

>>> video.length
355

    # display video metadata
    
>>> print video
Title: Rick Astley Sings Live - Never Gonna Give You Up - This Morning
Author: Ryan915
ID: cyMHZVT91Dw
Duration: 00:05:55
Rating: 4.93608852755
Views: 672583
Thumbnail: https://i1.ytimg.com/vi/cyMHZVT91Dw/default.jpg
Keywords: Rick, Astley, Sings, Live, on, This, Morning, Never, Gonna, Gunna, Give, You,...


    # show all formats for a video:
    
>>> streams = video.streams
>>> for s in streams:
>>>     print s.resolution, s.extension
480x854 webm
480x854 flv
360x640 webm
360x640 flv
360x640 mp4
240x400 flv
320x240 3gp
144x176 3gp


    # show all formats and their download url:

>>> for s in streams:
>>>     print s.resolution, s.extension, s.url
480x854 webm http://r12--sn-aoh8kier.c.youtube.com/videoplayback?expire=1369...
480x854 flv http://r11---sn-aoh8kier.c.youtube.com/videoplayback?expire=1369...
360x640 webm http://r11---sn-aoh8kier.c.youtube.com/videoplayback?expire=1369...
360x640 flv http://r11---sn-aoh8kier.c.youtube.com/videoplayback?expire=1369...
360x640 mp4 http://r11---sn-aoh8kier.c.youtube.com/videoplayback?expire=1369...
240x400 flv http://r11---sn-aoh8kier.c.youtube.com/videoplayback?expire=1369...
320x240 3gp http://r11---sn-aoh8kier.c.youtube.com/videoplayback?expire=1369...
144x176 3gp http://r11---sn-aoh8kier.c.youtube.com/videoplayback?expire=1369...


    # get best resolution regardless of file format:
    
>>> best = video.getbest()
>>> best.resolution, best.extension
('480x854', 'webm')


    # get best resolution for a particular file format:
    # (mp4, webm, flv or 3gp)
    
>>> best = video.getbest(preftype="mp4")
>>> best.resolution, best.extension
('360x640', 'mp4')


    # get best resolution for a particular file format, or return
    # a different format if it has the best resolution
    
>>> best = video.getbest(preftype="mp4", ftypestrict=False)
>>> best.resolution, best.extension
('480x854', 'webm')


    # get url, for download or streaming in mplayer / vlc etc
    
>>> best.url
'http://r12---sn-aig7kner.c.youtube.com/videoplayback?expire=1369...


    # Download video and show progress:
    
>>> best.download(progress=True)
-Downloading 'Rick Astley Sings Live - Never Gonna Give You Up - This Morning.webm' [56,858,674 Bytes]
Traceback (most recent call last):ed. Rate: [1095 kbps].  ETA: [33 secs]  


    # Download video, use specific filepath:
    
>>> myfilename = "/tmp/" + best.title + "." + best.extension
>>> best.download(progress=False, filepath=myfilename)
-Downloading 'Rick Astley Sings Live - Never Gonna Give You Up - This Morning.webm' [56,858,674 Bytes]
Done
```

Command Line Tool (ytdl) Usage:
===============================

```shell

usage: ytdl [-h] [-i] [-s] [-f {webm,mp4,3gp,flv}] [-r NNNxNNN] [-n N] [-b]
            url

YouTube Download Tool

positional arguments:
  url                   YouTube video URL to download

optional arguments:
  -h, --help            show this help message and exit
  -i                    Display video info
  -s                    Display available streams
  -n N                  Specify stream to download by stream number (use -s to
                        list, ignores -f and -r)
  -b                    Download the best quality video (ignores -n, -f and
                        -r)

format and quality:
  Specify the stream to download by resolution and file format

  -f {webm,mp4,3gp,flv}
                        format of the video to download
  -r NNNxNNN            resolution of the video to download

Examples:

# Download best available resolution (-b):

  ytdl "http://www.youtube.com/watch?v=cyMHZVT91Dw" -b

# get video info (-i):

  ytdl "http://www.youtube.com/watch?v=cyMHZVT91Dw" -i

# list available download formats (-s):

  ytdl "http://www.youtube.com/watch?v=cyMHZVT91Dw" -s
Stream  Format  Resolution  Size
------  ------  ----------  ----
0   webm    480x854      54 MB
1   flv     480x854      50 MB
2   webm    360x640      33 MB
3   flv     360x640      31 MB
4   mp4     360x640      24 MB
5   flv     240x400      13 MB
6   3gp     320x240      10 MB
7   3gp     144x176       3 MB

# Download mp4 360x640, ie. stream number 4 (-n4)

  ytdl "http://www.youtube.com/watch?v=cyMHZVT91Dw" -n 4

# Download flv at 240x400 (-f flv -r 240x400)
 
  ytdl "youtube.com/watch?v=cyMHZVT91Dw" -f flv -r 240x400 

```
"""
)
